import 'core/core.dart';
import 'core/keyset.dart';
import 'core/net/net.dart';

import 'net/net.dart';
import 'parser/parser.dart';

import 'dx/time.dart';
import 'dx/publish/publish.dart';
import 'dx/subscribe/subscribe.dart';
import 'dx/signal/signal.dart';
import 'dx/batch/batch.dart';
import 'dx/channel/channel.dart';
import 'dx/channel/channel_group.dart';
import 'dx/message_action/message_action.dart';
import 'dx/pam/pam.dart';
import 'dx/push/push.dart';
import 'dx/objects/schema.dart';
import 'dx/objects/membership.dart';
import 'dx/objects/space.dart';
import 'dx/objects/user.dart';
import 'dx/presence/presence.dart';

/// PubNub library.
///
/// All methods on this instance accept two named parameters: `keyset` and `using`.
/// * `keyset` accepts an instance of [Keyset],
/// * `using` accepts a String name of a keyset that was defined using `pubnub.keysets.add` method.
///
/// Both of those parameters are used to obtain an instance of [Keyset] that will be used in subsequent operations.
///
/// > **Example**: In case of method [channel], all operations on a returned [Channel] instance
/// > will be executed with a keyset that was obtained when the [channel] method was called.
///
///
/// At the beginning, the method that you call will try to obtain the keyset in accordance to those rules:
/// * `keyset` parameter has the highest priority,
/// * if `keyset` is null, it will try to obtain a keyset named `using`,
/// * if `using` is null, it will try to obtain the default keyset,
/// * if default keyset is not defined, it will throw an error.
class PubNub extends Core
    with
        TimeDx,
        PublishDx,
        SubscribeDx,
        SignalDx,
        MessageActionDx,
        PushNotificationDx,
        PamDx,
        PresenceDx {
  /// [BatchDx] contains methods that allow running batch operations on channels,
  /// channel groups and other features.
  BatchDx batch;

  /// [ChannelGroupDx] contains methods that allow manipulating channel groups.
  ChannelGroupDx channelGroups;

  /// [UserDx] contains methods that provide functionality to manage users.
  UserDx users;

  /// [SpaceDx] contains methods that allow manipulating spaces.
  SpaceDx spaces;

  /// [MembershipDx] contains methods that allow managing members of spaces and
  /// their users' memberships.
  MembershipDx memberships;

  /// Current version of this library.
  static String version = Core.version;

  PubNub({Keyset defaultKeyset, NetworkModule networking})
      : super(
            defaultKeyset: defaultKeyset,
            networking: networking ?? PubNubNetworkingModule(),
            parser: PubNubParserModule()) {
    batch = BatchDx(this);
    channelGroups = ChannelGroupDx(this);
    users = UserDx(this);
    spaces = SpaceDx(this);
    memberships = MembershipDx(this);
  }

  /// Returns a representation of a channel.
  ///
  /// Useful if you only need to work on one channel.
  Channel channel(String name, {Keyset keyset, String using}) {
    keyset ??= keysets.get(using, defaultIfNameIsNull: true);

    return Channel(this, keyset, name);
  }

  /// Returns a representation of a channel group.
  ///
  /// Useful if you need to work on a bunch of channels at the same time.
  ChannelGroup channelGroup(String name, {Keyset keyset, String using}) {
    keyset ??= keysets.get(using, defaultIfNameIsNull: true);

    return ChannelGroup(this, keyset, name);
  }

  /// Creates and returns a new instance of [User] (from Objects API).
  Future<User> user(String userId, String name,
      {String email,
      dynamic custom,
      String externalId,
      String profileUrl,
      Keyset keyset,
      String using}) async {
    keyset ??= keysets.get(using, defaultIfNameIsNull: true);

    User usr;
    var result = await users.create(
        UserDetails(userId, name,
            email: email,
            externalId: externalId,
            profileUrl: profileUrl,
            custom: custom),
        keyset: keyset);

    var userObject = result.data;
    if (result.status == 200) {
      usr = User(this, keyset, userObject.id);
    }
    return usr;
  }

  /// Creates and returns a new instance of [Space] (from Objects API).
  Future<Space> space(String spaceId, String name,
      {String description, dynamic custom, Keyset keyset, String using}) async {
    keyset ??= keysets.get(using, defaultIfNameIsNull: true);

    Space space;
    var result = await spaces.create(
        SpaceDetails(spaceId, name, description: description, custom: custom),
        keyset: keyset);

    var spaceObject = result.data;
    if (result.status == 200) {
      space = Space(this, keyset, spaceObject.id);
    }
    return space;
  }

  /// Returns a new instance of [Device] (from Push Notification API).
  ///
  /// [deviceId] should be non-empty and valid.
  Device device(String deviceId, {Keyset keyset, String using}) {
    keyset ??= keysets.get(using, defaultIfNameIsNull: true);
    return Device(this, keyset, deviceId);
  }
}
